@@ -23,6 +23,7 @@ messages about granted attribute access.
 Note that the ZOPE_WATCH_CHECKERS mechanism will eventually be
 replaced with a more general security auditing mechanism.
 """
+import abc
 import os
 import sys
 import types
@@ -30,26 +31,35 @@ import datetime
 import decimal
 import weakref
 
+from zope.i18nmessageid import Message
 import zope.interface.interface
 import zope.interface.interfaces
 import zope.interface.declarations
-from zope.i18nmessageid import Message
-from zope.interface import directlyProvides, Interface, implements
-from zope.interface.interfaces import IInterface, IDeclaration
+from zope.interface import Interface
+from zope.interface import directlyProvides
+from zope.interface import implementer
+from zope.interface.interfaces import IDeclaration
+from zope.interface.interfaces import IInterface
 
-from zope.security.interfaces import IChecker, INameBasedChecker
+from zope.security.interfaces import IChecker
+from zope.security.interfaces import INameBasedChecker
 from zope.security.interfaces import ISecurityProxyFactory
-from zope.security.interfaces import Unauthorized, ForbiddenAttribute
+from zope.security.interfaces import ForbiddenAttribute
+from zope.security.interfaces import Unauthorized
 from zope.security._definitions import thread_local
-from zope.security._proxy import _Proxy as Proxy, getChecker
+from zope.security._compat import CLASS_TYPES
+from zope.security._compat import PYTHON2
+from zope.security._compat import _u
+from zope.security.proxy import Proxy
+from zope.security.proxy import getChecker
 
 try:
     from zope.exceptions import DuplicationError
-except ImportError:
+except ImportError: #pragma NO COVER
     class DuplicationError(Exception):
         """A duplicate registration was attempted"""
 
-if os.environ.get('ZOPE_WATCH_CHECKERS'):
+if os.environ.get('ZOPE_WATCH_CHECKERS'): #pragma NO COVER
     try:
         WATCH_CHECKERS = int(os.environ.get('ZOPE_WATCH_CHECKERS'))
     except ValueError:
@@ -88,6 +98,10 @@ def ProxyFactory(object, checker=None):
 
 directlyProvides(ProxyFactory, ISecurityProxyFactory)
 
+# This import represents part of the API for the proxy module
+from . import proxy
+proxy.ProxyFactory = ProxyFactory
+
 def canWrite(obj, name):
     """Check whether the interaction may write an attribute named name on obj.
 
@@ -140,8 +154,8 @@ def canAccess(obj, name):
     # indicates a programming or configuration error
     return True
 
-class Checker(object):
-    implements(INameBasedChecker)
+@implementer(INameBasedChecker)
+class CheckerPy(object):
 
     def __init__(self, get_permissions, set_permissions=None):
         """Create a checker
@@ -157,10 +171,14 @@ class Checker(object):
         set attribute access.
 
         """
-        assert isinstance(get_permissions, dict)
+        if not isinstance(get_permissions, dict):
+            raise TypeError('get_permissions must be a dict')
         self.get_permissions = get_permissions
         if set_permissions is not None:
-            assert isinstance(set_permissions, dict)
+            if not isinstance(set_permissions, dict):
+                raise TypeError('set_permissions must be a dict')
+        else:
+            set_permissions = {}
         self.set_permissions = set_permissions
 
     def permission_id(self, name):
@@ -172,9 +190,6 @@ class Checker(object):
         if self.set_permissions:
             return self.set_permissions.get(name)
 
-    def check_getattr(self, object, name):
-        'See IChecker'
-        self.check(object, name)
 
     def check_setattr(self, object, name):
         'See IChecker'
@@ -213,6 +228,8 @@ class Checker(object):
             __traceback_supplement__ = (TracebackSupplement, object)
             raise ForbiddenAttribute(name, object)
 
+    check_getattr = check # 'See IChecker'
+
     def proxy(self, value):
         'See IChecker'
         if type(value) is Proxy:
@@ -224,7 +241,7 @@ class Checker(object):
                 return value
 
         return Proxy(value, checker)
-
+Checker = CheckerPy # in case no C optimizations
 
 
 # Helper class for __traceback_supplement__
@@ -239,19 +256,19 @@ class TracebackSupplement(object):
             cls = self.obj.__class__
             if hasattr(cls, "__module__"):
                 s = "%s.%s" % (cls.__module__, cls.__name__)
-            else:
+            else: #pragma NO COVER XXX
                 s = str(cls.__name__)
             result.append("   - class: " + s)
-        except:
+        except: #pragma NO COVER XXX
             pass
         try:
             cls = type(self.obj)
             if hasattr(cls, "__module__"):
                 s = "%s.%s" % (cls.__module__, cls.__name__)
-            else:
+            else: #pragma NO COVER XXX
                 s = str(cls.__name__)
             result.append("   - type: " + s)
-        except:
+        except: #pragma NO COVER XXX
             pass
         return "\n".join(result)
 
@@ -265,7 +282,7 @@ class Global(object):
     """
 
     def __init__(self, name, module=None):
-        if module is None:
+        if module is None: #pragma NO COVER XXX
             module = sys._getframe(1).f_locals['__name__']
 
         self.__name__ = name
@@ -280,12 +297,14 @@ class Global(object):
 
 # Marker for public attributes
 CheckerPublic = Global('CheckerPublic')
+CP_HACK_XXX = CheckerPublic
 
-# Now we wrap it in a security proxy so that it retains it's
+# Now we wrap it in a security proxy so that it retains its
 # identity when it needs to be security proxied.
-d={}
-CheckerPublic = Proxy(CheckerPublic, Checker(d))
+d = {}
+CheckerPublic = Proxy(CheckerPublic, Checker(d)) # XXX uses CheckerPy
 d['__reduce__'] = CheckerPublic
+d['__module__'] = CheckerPublic
 del d
 
 # TODO: It's a bit scary above that we can pickle a proxy if access is
@@ -349,7 +368,7 @@ def MultiChecker(specs):
 
     return Checker(data)
 
-def selectChecker(object):
+def selectCheckerPy(object):
     """Get a checker for the given object
 
     The appropriate checker is returned or None is returned. If the
@@ -380,17 +399,19 @@ def selectChecker(object):
             return None
 
     return checker
+selectChecker = selectCheckerPy # in case no C optimizations
 
 def getCheckerForInstancesOf(class_):
     return _checkers.get(class_)
 
+DEFINABLE_TYPES = CLASS_TYPES + (types.ModuleType,)
 def defineChecker(type_, checker):
     """Define a checker for a given type of object
 
     The checker can be a Checker, or a function that, when called with
     an object, returns a Checker.
     """
-    if not isinstance(type_, (type, types.ClassType, types.ModuleType)):
+    if not isinstance(type_, DEFINABLE_TYPES):
         raise TypeError(
                 'type_ must be a type, class or module, not a %s' % type_)
     if type_ in _checkers:
@@ -420,7 +441,7 @@ _available_by_default = []
 # Get optimized versions
 try:
     import zope.security._zope_security_checker
-except ImportError:
+except ImportError: #pragma NO COVER
     pass
 else:
     from zope.security._zope_security_checker import _checkers, selectChecker
@@ -432,6 +453,7 @@ else:
 
 _getChecker = _checkers.get
 
+@implementer(IChecker)
 class CombinedChecker(Checker):
     """A checker that combines two other checkers in a logical-or fashion.
 
@@ -448,7 +470,6 @@ class CombinedChecker(Checker):
     ForbiddenAttribute ForbiddenAttribute ForbiddenAttribute
     ------------------ ------------------ -----------------------------------
     """
-    implements(IChecker)
 
     def __init__(self, checker1, checker2):
         """Create a combined checker."""
@@ -463,7 +484,7 @@ class CombinedChecker(Checker):
             Checker.check(self, object, name)
         except ForbiddenAttribute:
             self._checker2.check(object, name)
-        except Unauthorized, unauthorized_exception:
+        except Unauthorized as unauthorized_exception:
             try: self._checker2.check(object, name)
             except ForbiddenAttribute:
                 raise unauthorized_exception
@@ -476,7 +497,7 @@ class CombinedChecker(Checker):
             Checker.check_setattr(self, object, name)
         except ForbiddenAttribute:
             self._checker2.check_setattr(object, name)
-        except Unauthorized, unauthorized_exception:
+        except Unauthorized as unauthorized_exception:
             try: self._checker2.check_setattr(object, name)
             except ForbiddenAttribute:
                 raise unauthorized_exception
@@ -492,23 +513,28 @@ class CheckerLoggingMixin(object):
     """
 
     verbosity = 1
+    _file = sys.stderr
+
+    def _log(self, msg, verbosity=1):
+        if self.verbosity >= verbosity:
+            self._file.write('%s\n' % msg)
 
     def check(self, object, name):
         try:
             super(CheckerLoggingMixin, self).check(object, name)
             if self.verbosity > 1:
                 if name in _available_by_default:
-                    print >> sys.stderr, (
-                        '[CHK] + Always available: %s on %r' % (name, object))
+                    self._log('[CHK] + Always available: %s on %r'
+                        % (name, object), 2)
                 else:
-                    print >> sys.stderr, (
-                        '[CHK] + Granted: %s on %r' % (name, object))
+                    self._log(
+                        '[CHK] + Granted: %s on %r' % (name, object), 2)
         except Unauthorized:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Unauthorized: %s on %r' % (name, object))
             raise
         except ForbiddenAttribute:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Forbidden: %s on %r' % (name, object))
             raise
 
@@ -517,39 +543,41 @@ class CheckerLoggingMixin(object):
             super(CheckerLoggingMixin, self).check(object, name)
             if self.verbosity > 1:
                 if name in _available_by_default:
-                    print >> sys.stderr, (
+                    self._log(
                         '[CHK] + Always available getattr: %s on %r'
-                        % (name, object))
+                        % (name, object), 2)
                 else:
-                    print >> sys.stderr, (
+                    self._log(
                         '[CHK] + Granted getattr: %s on %r'
-                        % (name, object))
+                        % (name, object), 2)
         except Unauthorized:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Unauthorized getattr: %s on %r' % (name, object))
             raise
         except ForbiddenAttribute:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Forbidden getattr: %s on %r' % (name, object))
             raise
 
+    __setitem__ = check_getattr
+
     def check_setattr(self, object, name):
         try:
             super(CheckerLoggingMixin, self).check_setattr(object, name)
             if self.verbosity > 1:
-                print >> sys.stderr, (
-                    '[CHK] + Granted setattr: %s on %r' % (name, object))
+                self._log(
+                    '[CHK] + Granted setattr: %s on %r' % (name, object), 2)
         except Unauthorized:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Unauthorized setattr: %s on %r' % (name, object))
             raise
         except ForbiddenAttribute:
-            print >> sys.stderr, (
+            self._log(
                 '[CHK] - Forbidden setattr: %s on %r' % (name, object))
             raise
 
 
-if WATCH_CHECKERS:
+if WATCH_CHECKERS: #pragma NO COVER
     class Checker(CheckerLoggingMixin, Checker):
         verbosity = WATCH_CHECKERS
     class CombinedChecker(CheckerLoggingMixin, CombinedChecker):
@@ -576,7 +604,7 @@ _typeChecker = NamesChecker(
      '__implemented__'])
 _namedChecker = NamesChecker(['__name__'])
 
-_iteratorChecker = NamesChecker(['next', '__iter__'])
+_iteratorChecker = NamesChecker(['next', '__iter__', '__len__'])
 
 _setChecker = NamesChecker(['__iter__', '__len__', '__str__', '__contains__',
                             'copy', 'difference', 'intersection', 'issubset',
@@ -611,11 +639,10 @@ _basic_types = {
     object: NoProxy,
     int: NoProxy,
     float: NoProxy,
-    long: NoProxy,
     complex: NoProxy,
-    types.NoneType: NoProxy,
+    type(None): NoProxy,
     str: NoProxy,
-    unicode: NoProxy,
+    bytes: NoProxy,
     Message: NoProxy, # Messages are immutable, so it's okay
     bool: NoProxy,
     datetime.timedelta: NoProxy,
@@ -624,11 +651,19 @@ _basic_types = {
     datetime.time: NoProxy,
     datetime.tzinfo: NoProxy,
 }
+if PYTHON2: 
+    _basic_types[long] = NoProxy
+    _basic_types[unicode] = NoProxy
+else: #pragma NO COVER
+    _basic_types[type({}.values())] = NoProxy
+    _basic_types[type({}.keys())] = NoProxy
+    _basic_types[type({}.items())] = NoProxy
+
 try:
     import pytz
 except ImportError:
     pass
-else:
+else: #pragma NO COVER
     _basic_types[type(pytz.UTC)] = NoProxy
 
 BasicTypes = BasicTypes(_basic_types)
@@ -639,11 +674,9 @@ BasicTypes_examples = {
     object: object(),
     int: 65536,
     float: -1.4142,
-    long: 65536l,
     complex: -1.4142j,
-    types.NoneType: None,
-    str: 'abc',
-    unicode: u'uabc',
+    type(None): None,
+    bytes: b'abc',
     bool: True,
     datetime.timedelta: datetime.timedelta(3),
     datetime.datetime: datetime.datetime(2003, 1, 1),
@@ -652,8 +685,12 @@ BasicTypes_examples = {
     Message: Message('message', domain='hello')
 }
 
+if PYTHON2:
+    BasicTypes_examples[unicode] = _u('uabc')
+    BasicTypes_examples[long] = long(65536)
 
-class _Sequence(object):
+
+class _Sequence(object): #pragma NO COVER
     def __len__(self): return 0
     def __getitem__(self, i): raise IndexError
 
@@ -665,7 +702,7 @@ _Declaration_checker = InterfaceChecker(
     __call__=CheckerPublic,
     )
 
-def f():
+def f(): #pragma NO COVER
     yield f
 
 
@@ -698,24 +735,18 @@ _default_checkers = {
     tuple: NamesChecker(['__getitem__', '__getslice__', '__add__', '__radd__',
                          '__contains__', '__len__', '__iter__',
                          '__str__']),
-    types.InstanceType: _instanceChecker,
     Proxy: NoProxy,
     type(weakref.ref(_Sequence())): NamesChecker(['__call__']),
-    types.ClassType: _typeChecker,
     types.FunctionType: _callableChecker,
     types.MethodType: _callableChecker,
     types.BuiltinFunctionType: _callableChecker,
     types.BuiltinMethodType: _callableChecker,
-    type(().__getslice__): _callableChecker, # slot description
     type: _typeChecker,
     types.ModuleType: lambda module: _checkers.get(module, _namedChecker),
     type(iter([])): _iteratorChecker, # Same types in Python 2.2.1,
     type(iter(())): _iteratorChecker, # different in Python 2.3.
     type(iter({})): _iteratorChecker,
     type(iter(set())): _iteratorChecker,
-    type({}.iteritems()): _iteratorChecker,
-    type({}.iterkeys()): _iteratorChecker,
-    type({}.itervalues()): _iteratorChecker,
     type(iter(_Sequence())): _iteratorChecker,
     type(f()): _iteratorChecker,
     type(Interface): InterfaceChecker(
@@ -728,17 +759,16 @@ _default_checkers = {
     zope.interface.declarations.ClassProvides: _Declaration_checker,
     zope.interface.declarations.Implements: _Declaration_checker,
     zope.interface.declarations.Declaration: _Declaration_checker,
+    abc.ABCMeta: _typeChecker,
 }
-
-if sys.version_info < (2, 6):
-    import sets
-    _default_checkers[sets.Set] = _setChecker
-    _default_checkers[sets.ImmutableSet] = _setChecker
-
-if sys.version_info >= (2, 6):
-    import abc
-    _default_checkers[abc.ABCMeta] = _typeChecker
-
+if PYTHON2:
+    _default_checkers[types.ClassType] = _typeChecker
+    _default_checkers[types.InstanceType] = _instanceChecker
+    # slot description
+    _default_checkers[type(().__getslice__)] = _callableChecker
+    _default_checkers[type({}.iteritems())] = _iteratorChecker
+    _default_checkers[type({}.iterkeys())] = _iteratorChecker
+    _default_checkers[type({}.itervalues())] = _iteratorChecker
 
 def _clear():
     _checkers.clear()
@@ -749,7 +779,7 @@ _clear()
 
 try:
     from zope.testing.cleanup import addCleanUp
-except ImportError:
+except ImportError: #pragma NO COVER
     pass
 else:
     addCleanUp(_clear)
